"""
    cutitout: automatically cut silence from videos
    Copyright (C) 2020 Wolf Clement

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import audioop
import subprocess
import sys
from typing import Dict, List, Tuple


# Returns the index of the first audio stream and its sample rate.
def get_sample_rate(filename) -> Tuple[int, int]:
    streams: List[Dict] = []
    probe = subprocess.check_output(
        ["ffprobe", "-show_streams", filename],
        encoding="utf-8",
        stderr=subprocess.DEVNULL,
    )

    for line in probe.split("\n"):
        if line == "[STREAM]":
            streams.append({})

        try:
            key, value = line.split("=")
            streams[-1][key] = value
        except ValueError:
            pass

    audio_stream = next(filter(lambda s: s["codec_type"] == "audio", streams))
    return int(audio_stream["index"]), int(audio_stream["sample_rate"])


if __name__ == "__main__":
    filename = sys.argv[1]
    minute = sys.argv[2]

    # Values configurable in cutitout.lua
    clip_margin = float(sys.argv[3])
    assert clip_margin >= 0.0
    audio_treshold = float(sys.argv[4])
    assert audio_treshold > 0.0 and audio_treshold <= 1.0
    min_clip_length = float(sys.argv[5])
    assert min_clip_length > 0.0
    min_skip_length = float(sys.argv[6])
    assert min_skip_length > 2 * clip_margin

    # Offset in seconds, at which we start searching for skips
    start_offset = float(sys.argv[2]) * 60.0

    # hh:mm:ss.xx formatted start offset to pass to ffmpeg
    hours = int(float(sys.argv[2]) / 60)
    minutes = int(sys.argv[2]) % 60
    start_str = f"{hours}:{minutes}:00.00"

    # Set the fragment size for 10ms long audio fragments (* 2 because we get 2 bytes)
    index, sample_rate = get_sample_rate(filename)
    fragment_length = int(sample_rate * 0.01 * 2)

    # Start an ffmpeg process that will pipe us raw audio fragments
    orig_audio = subprocess.Popen(
        [
            "ffmpeg",
            "-ss",
            start_str,
            "-t",
            "1:00.00",
            "-i",
            filename,
            # Output only one channel
            "-ac",
            "1",
            # Output raw 16bit samples for fast processing
            "-f",
            "s16le",
            # Open specific audio stream
            "-map",
            f"0:{index}",
            # Only use one core to avoid making mpv lag
            "-threads",
            "1",
            # Pipe to orig_audio
            "pipe:1",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    )
    if orig_audio.stdout is None:
        raise Exception("Failed to start ffmpeg. Is it installed?")

    clips: List[Tuple[float, float]] = []
    clip_index = 0
    loud_start = -1

    chunk_data = orig_audio.stdout.read(fragment_length)
    while chunk_data:
        # With *signed* 16 bit audio, the maximal absolute value is 2^15 = 32768.
        volume = audioop.max(chunk_data, 2) / 32768

        if loud_start == -1 and volume >= audio_treshold:
            loud_start = clip_index
        elif loud_start != -1 and volume < audio_treshold:
            # Remove sounds that are too short to be important
            if clip_index - loud_start > min_clip_length * 100:
                clips.append((loud_start / 100, clip_index / 100))
            loud_start = -1

        chunk_data = orig_audio.stdout.read(fragment_length)
        clip_index += 1

    # Turn clips into skips
    skips: List[Tuple[float, float]] = []
    current_time = 0.0
    for clip in clips:
        clip_start = clip[0]
        clip_end = clip[1]

        if clip_start - current_time < min_skip_length:
            current_time = clip_end + clip_margin
        else:
            skips.append((current_time, clip_start - clip_margin))
            current_time = clip_end + clip_margin

    if 60.0 - current_time >= min_skip_length:
        skips.append((current_time, 60.0))

    skips = [(start_offset + s[0], start_offset + s[1]) for s in skips]
    skip_table = ["{" + f"{s[0]},{s[1]}" + "}" for s in skips]
    print("return {" + ",".join(skip_table) + "}")
